from unittest import TestCase, mock
import random
import string
import os
import asyncio
import pytest
from asynctest import CoroutineMock

import service
from simulation.worker_managers.local_worker_manager import LocalWorkerManager
from simulation.game_runner import GameRunner


class MockGameState(object):
    def serialise(self):
        return {'foo': 'bar'}


class MockedSocketIOServer(mock.MagicMock):
    """ Decorator function that just returns the function. Needed because we decorate
        functions in the GameAPI."""
    def on(self, event):
        def decorator(func):
            def wrapper(*args, **kwargs):
                func(*args, **kwargs)
            return wrapper
        return decorator

class TestSocketIO:

    def setup_method(self, method):
        os.environ['GAME_ID'] = '1'
        self.environ = {'QUERY_STRING': 'avatar_id=1&EIO=3&transport=polling&t=MJhoMgb'}
        self.game_api = self.create_game_api()
        self.mocked_mappings = self.game_api._socket_session_id_to_player_id
        self.sid = ''.join(random.choice(string.ascii_uppercase +
                                         string.ascii_lowercase +
                                         string.digits)
                           for _ in range(19))

    def teardown_method(self, method):
        del os.environ['GAME_ID']

    @mock.patch('docker.from_env')
    @mock.patch('service.app')
    def create_game_api(self, app, docker_from_env):
        game_runner = GameRunner(worker_manager_class=LocalWorkerManager,
                                 game_state_generator=lambda avatar_manager: MockGameState(),
                                 django_api_url='http://test',
                                 port='0000')
        return service.GameAPI(game_state=game_runner.game_state,
                               worker_manager=game_runner.worker_manager)

    @pytest.mark.asyncio
    @mock.patch('service.app')
    @mock.patch('service.socketio_server', new_callable=MockedSocketIOServer)
    async def test_socketio_emit_called(self, mocked_socketio, app):
        self.game_api.worker_manager.add_new_worker(1)
        await self.game_api.register_world_update_on_connect()(self.sid, self.environ)

        assert mocked_socketio.manager.emit.mockreturn_value.emit.assert_called_once

    @pytest.mark.asyncio
    @mock.patch('service.app')
    @mock.patch('service.socketio_server.emit', new_callable=CoroutineMock())
    @mock.patch('service.socketio_server', new_callable=MockedSocketIOServer)
    async def test_matched_session_id_to_avatar_id_mapping(self, mocked_socketio, mocked_emit, app):
        assert len(self.mocked_mappings) == 0

        self.game_api.worker_manager.add_new_worker(1)
        await self.game_api.register_world_update_on_connect()(self.sid, self.environ)

        assert len(self.mocked_mappings) == 1
        assert self.sid in self.mocked_mappings
        assert int(self.mocked_mappings[self.sid]) == 1

    @pytest.mark.asyncio
    @mock.patch('service.app')
    @mock.patch('service.socketio_server', new_callable=MockedSocketIOServer)
    async def test_no_match_session_id_to_avatar_id_mapping(self, mocked_socketio, app):
        self.environ['QUERY_STRING'] = 'corrupted!@$%string123'

        assert len(self.mocked_mappings) ==  0

        self.game_api.worker_manager.add_new_worker(1)
        await self.game_api.register_world_update_on_connect()(self.sid, self.environ)

        assert len(self.mocked_mappings) ==  0
        assert not self.sid in self.mocked_mappings

    @pytest.mark.asyncio
    @mock.patch('service.app')
    @mock.patch('service.socketio_server', new_callable=MockedSocketIOServer)
    async def test_send_updates_for_one_user(self, mocked_socketio, app):
        self.mocked_mappings[self.sid] = 1
        self.game_api.worker_manager.add_new_worker(self.mocked_mappings[self.sid])
        worker = self.game_api.worker_manager.player_id_to_worker[self.mocked_mappings[self.sid]]
        worker.log = 'Logs one'

        with mock.patch('service.socketio_server.emit', new=CoroutineMock()) as mocked_emit:
            await self.game_api.send_updates()

            game_state_call = mock.call('game-state', {'foo': 'bar'}, room=self.sid)
            log_call = mock.call('log', 'Logs one', room=self.sid)

            mocked_emit.assert_has_calls([game_state_call, log_call], any_order=True)

    @pytest.mark.asyncio
    @mock.patch('service.app')
    @mock.patch('service.socketio_server', new_callable=MockedSocketIOServer)
    async def test_no_logs_not_emitted(self, mocked_socketio, app):
        """ If there are no logs for an avatar, no logs should be emitted. """
        self.mocked_mappings[self.sid] = 1
        self.game_api.worker_manager.add_new_worker(self.mocked_mappings[self.sid])

        with mock.patch('service.socketio_server.emit', new=CoroutineMock()) as mocked_emit:
            await self.game_api.send_updates()

            mocked_emit.assert_called_once_with('game-state', {'foo': 'bar'}, room=self.sid)

    @pytest.mark.asyncio
    @mock.patch('service.app')
    @mock.patch('service.socketio_server', new_callable=MockedSocketIOServer)
    async def test_empty_logs_not_emitted(self, mocked_socketio, app):
        """ If the logs are an empty sting, no logs should be emitted. """
        self.mocked_mappings[self.sid] = 1

        self.game_api.worker_manager.add_new_worker(self.mocked_mappings[self.sid])
        worker = self.game_api.worker_manager.player_id_to_worker[self.mocked_mappings[self.sid]]
        worker.logs = ''
        
        with mock.patch('service.socketio_server.emit', new=CoroutineMock()) as mocked_emit:
            await self.game_api.send_updates()

            mocked_emit.assert_called_once_with('game-state', {'foo': 'bar'}, room=self.sid)

    @pytest.mark.asyncio
    @mock.patch('service.app')
    @mock.patch('service.socketio_server', new_callable=MockedSocketIOServer)
    async def test_send_updates_for_multiple_users(self, mocked_socketio, app):
        self.mocked_mappings[self.sid] = 1
        self.mocked_mappings['differentsid'] = 2

        self.game_api.worker_manager.add_new_worker(self.mocked_mappings[self.sid])
        self.game_api.worker_manager.add_new_worker(self.mocked_mappings['differentsid'])
        worker_one = self.game_api.worker_manager.player_id_to_worker[self.mocked_mappings[self.sid]]
        worker_two = self.game_api.worker_manager.player_id_to_worker[self.mocked_mappings['differentsid']]
        worker_one.log = 'Logs one'
        worker_two.log = 'Logs two'

        with mock.patch('service.socketio_server.emit', new=CoroutineMock()) as mocked_emit:
            await self.game_api.send_updates()

            user_one_game_state_call = mock.call('game-state', {'foo': 'bar'}, room=self.sid)
            user_two_game_state_call = mock.call('game-state', {'foo': 'bar'}, room='differentsid')
            user_one_log_call = mock.call('log', 'Logs one', room=self.sid)
            user_two_log_call = mock.call('log', 'Logs two', room='differentsid')

            mocked_emit.assert_has_calls([user_one_game_state_call,
                                                   user_two_game_state_call,
                                                   user_one_log_call,
                                                   user_two_log_call], any_order=True)

    @pytest.mark.asyncio
    @mock.patch('service.app')
    @mock.patch('service.socketio_server', new_callable=MockedSocketIOServer)
    async def test_send_code_changed_flag(self, mocked_socketio, app):
        self.mocked_mappings[self.sid] = 1
        self.game_api.worker_manager.add_new_worker(self.mocked_mappings[self.sid])
        worker = self.game_api.worker_manager.player_id_to_worker[self.mocked_mappings[self.sid]]
        worker.has_code_updated = True
        
        with mock.patch('service.socketio_server.emit', new=CoroutineMock()) as mocked_emit:        
            await self.game_api.send_updates()

            user_game_state_call = mock.call('game-state', {'foo': 'bar'}, room=self.sid)
            user_game_code_changed_call = mock.call('feedback-avatar-updated', {}, room=self.sid)

            mocked_emit.assert_has_calls([user_game_state_call, user_game_code_changed_call], any_order=True)

    @pytest.mark.asyncio
    @mock.patch('service.app')
    @mock.patch('service.socketio_server', new_callable=MockedSocketIOServer)
    async def test_send_false_flag_not_sent(self, mocked_socketio, app):
        self.mocked_mappings[self.sid] = 1
        self.game_api.worker_manager.add_new_worker(self.mocked_mappings[self.sid])
        worker = self.game_api.worker_manager.player_id_to_worker[self.mocked_mappings[self.sid]]
        worker.has_code_updated = False

        with mock.patch('service.socketio_server.emit', new=CoroutineMock()) as mocked_emit:
            await self.game_api.send_updates()

            mocked_emit.assert_called_once_with('game-state', {'foo': 'bar'}, room=self.sid)

    @pytest.mark.asyncio
    async def test_remove_session_id_on_disconnect(self):
        self.mocked_mappings[self.sid] = 1
        assert len(self.mocked_mappings) ==  1
        assert self.sid in self.mocked_mappings
        assert self.mocked_mappings[self.sid] ==  1

        await self.game_api.register_remove_session_id_from_mappings()(sid=self.sid)

        assert self.sid not in self.mocked_mappings
        assert len(self.mocked_mappings) ==  0



import atexit
import itertools
import logging
import os
import subprocess
import docker
import json

from .worker_manager import WorkerManager

LOGGER = logging.getLogger(__name__)


class LocalWorkerManager(WorkerManager):
    """Relies on them already being created already."""

    host = os.environ.get('LOCALHOST_IP', '127.0.0.1')
    worker_directory = os.path.join(
        os.path.dirname(__file__),
        '../../../aimmo-game-worker/',
    )

    def __init__(self, *args, **kwargs):
        self.workers = {}
        self.game_id = os.environ['GAME_ID']
        self.port_counter = itertools.count(1989 + int(self.game_id) * 10000)
        self.client = docker.from_env()
        super(LocalWorkerManager, self).__init__(*args, **kwargs)

    def create_worker(self, player_id):
        assert(player_id not in self.workers)
        port = next(self.port_counter)

        template_string = os.environ.get('CONTAINER_TEMPLATE')
        if template_string:
            template = json.loads(template_string)
        else:
            template = {
                'environment': {}
            }
        data_url = 'http://{}:{}/player/{}'.format(self.host, self.port, player_id)
        template['environment']['DATA_URL'] = data_url
        template['environment']['PORT'] = port
        container = self.client.containers.run(
            name="aimmo-{}-worker-{}".format(self.game_id, player_id),
            image='ocadotechnology/aimmo-game-worker:test',
            ports={f"{port}/tcp": port},
            **template)
        self.workers[player_id] = container
        worker_url = 'http://%s:%d' % (
            self.host,
            port,
        )
        LOGGER.info("Worker started for %s, listening at %s", player_id, worker_url)
        return worker_url

    def remove_worker(self, player_id):
        if player_id in self.workers:
            self.workers[player_id].kill()
            del self.workers[player_id]



#!/usr/bin/env python

from aiohttp import web
import aiohttp_cors
import asyncio
import pickle
import ast
import logging
import os
import sys
import json
from urllib.parse import parse_qs

import socketio

from simulation import map_generator
from simulation.worker_managers import WORKER_MANAGERS
from simulation.game_runner import GameRunner

app = web.Application()
cors = aiohttp_cors.setup(app)

socketio_server = socketio.AsyncServer(async_handlers=True)

routes = web.RouteTableDef()

LOGGER = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


class GameAPI(object):
    def __init__(self, game_state, worker_manager):
        self._socket_session_id_to_player_id = {}
        self.register_endpoints()
        self.worker_manager = worker_manager
        self.game_state = game_state

    def register_endpoints(self):
        self.register_player_data_view()
        self.register_world_update_on_connect()
        self.register_remove_session_id_from_mappings()
        self.register_healthcheck()
        app.add_routes(routes)

    def register_healthcheck(self):
        @routes.get('/game-{game_id}')
        async def healthcheck(request):
            return web.Response(text='HEALTHY')

        return healthcheck

    def register_player_data_view(self):
        @routes.get('/player/{player_id}')
        async def player_data(request: web.Request):
            player_id = int(request.match_info['player_id'])
            return web.json_response({
                'code': self.worker_manager.get_code(player_id),
                'options': {},
                'state': None,
            })

        return player_data

    def register_world_update_on_connect(self):
        @socketio_server.on('connect')
        async def world_update_on_connect(sid, environ):
            query = environ['QUERY_STRING']
            self._find_avatar_id_from_query(sid, query)
            await self.send_updates()

        return world_update_on_connect

    def register_remove_session_id_from_mappings(self):
        @socketio_server.on('disconnect')
        async def remove_session_id_from_mappings(sid):
            LOGGER.info("Socket disconnected for session id:{}. ".format(sid))
            try:
                del self._socket_session_id_to_player_id[sid]
            except KeyError:
                pass

        return remove_session_id_from_mappings

    async def send_updates(self):
        player_id_to_worker = self.worker_manager.player_id_to_worker
        await self._send_have_avatars_code_updated(player_id_to_worker)
        await self._send_game_state()
        await self._send_logs(player_id_to_worker)

    def _find_avatar_id_from_query(self, session_id, query_string):
        """
        :param session_id: Int with the session id
        :param query_string: String from the environment settings,
        usually located as the key 'QUERY_STRING'.
        """
        parsed_qs = parse_qs(query_string)

        try:
            avatar_id = int(parsed_qs['avatar_id'][0])
            self._socket_session_id_to_player_id[session_id] = avatar_id
        except ValueError:
            LOGGER.error("Avatar ID could not be casted into an integer")
        except KeyError:
            LOGGER.error("No avatar ID found. User may not be authorised ")
            LOGGER.error("query_string: " + query_string)

    async def _send_logs(self, player_id_to_workers):
        def should_send_logs(logs):
            return bool(logs)

        socket_session_id_to_player_id_copy = self._socket_session_id_to_player_id.copy()
        for sid, player_id in socket_session_id_to_player_id_copy.items():
            avatar_logs = player_id_to_workers[player_id].log
            if should_send_logs(avatar_logs):
                await socketio_server.emit('log', avatar_logs, room=sid)

    async def _send_game_state(self):
        serialised_game_state = self.game_state.serialise()
        socket_session_id_to_player_id_copy = self._socket_session_id_to_player_id.copy()
        for sid, player_id in socket_session_id_to_player_id_copy.items():
            await socketio_server.emit('game-state', serialised_game_state, room=sid)

    async def _send_have_avatars_code_updated(self, player_id_to_workers):
        socket_session_id_to_player_id_copy = self._socket_session_id_to_player_id.copy()
        for sid, player_id in socket_session_id_to_player_id_copy.items():
            if player_id_to_workers[player_id].has_code_updated:
                await socketio_server.emit('feedback-avatar-updated', {}, room=sid)


def create_runner(port):
    settings = json.loads(os.environ['settings'])
    generator = getattr(map_generator, settings['GENERATOR'])(settings)
    worker_manager_class = WORKER_MANAGERS[os.environ.get('WORKER_MANAGER', 'local')]
    return GameRunner(worker_manager_class=worker_manager_class,
                      game_state_generator=generator.get_game_state,
                      django_api_url=os.environ.get('GAME_API_URL', 'http://localhost:8000/aimmo/api/games/'),
                      port=port)


def run_game(port):
    game_runner = create_runner(port)
    game_api = GameAPI(game_state=game_runner.game_state,
                       worker_manager=game_runner.worker_manager)
    game_runner.set_end_turn_callback(game_api.send_updates)
    asyncio.ensure_future(game_runner.run())


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    host = sys.argv[1]

    socketio_server.attach(app, socketio_path=os.environ.get('SOCKETIO_RESOURCE', 'socket.io'))

    if os.environ['WORKER_MANAGER'] == 'local':
        port = int(os.environ['EXTERNAL_PORT'])
    else:
        port = int(sys.argv[2])

    run_game(port)

    LOGGER.info("starting the server")
    web.run_app(app, host=host, port=port)



from django.conf import settings
from django.utils.module_loading import import_string
from permissions import default_preview_user
from django.contrib.auth.models import User

#: URL function for locating the game server, takes one parameter `game`
GAME_SERVER_URL_FUNCTION = getattr(settings, 'AIMMO_GAME_SERVER_URL_FUNCTION', None)
GAME_SERVER_PORT_FUNCTION = getattr(settings, 'AIMMO_GAME_SERVER_PORT_FUNCTION', None)
GAME_SERVER_SSL_FLAG = getattr(settings, 'AIMMO_GAME_SERVER_SSL_FLAG', False)
PREVIEW_USER_AIMMO_DECORATOR = getattr(settings, 'PREVIEW_USER_AIMMO_DECORATOR', None)
USERS_FOR_NEW_AIMMO_GAME = getattr(settings, 'USERS_FOR_NEW_AIMMO_GAME', None)


def get_aimmo_preview_user_decorator():
    """
    This function is used to import a decorator from portal, which
    checks whether the logged in user is a preview user.

    :return: A function decorator
    """
    if PREVIEW_USER_AIMMO_DECORATOR:
        func = import_string(PREVIEW_USER_AIMMO_DECORATOR)
        return func
    return default_preview_user


def get_users_for_new_game(request):
    """
    Imports and calls a function defined in portal, that decides
    which users should be added to the newly created game.
    :param request:
    :return: List of User objects
    """
    if USERS_FOR_NEW_AIMMO_GAME:
        func = import_string(USERS_FOR_NEW_AIMMO_GAME)
        return func(request)
    return User.objects.all()


preview_user_required = get_aimmo_preview_user_decorator()

MAX_LEVEL = 1



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('aimmo', '0006_game_static_data'),
    ]

    operations = [
        migrations.AlterField(
            model_name='game',
            name='pickup_spawn_chance',
            field=models.FloatField(default=0.1),
        ),
        migrations.AlterField(
            model_name='game',
            name='score_despawn_chance',
            field=models.FloatField(default=0.05),
        ),
        migrations.AlterField(
            model_name='game',
            name='start_height',
            field=models.IntegerField(default=31),
        ),
        migrations.AlterField(
            model_name='game',
            name='start_width',
            field=models.IntegerField(default=31),
        ),
        migrations.AlterField(
            model_name='game',
            name='target_num_pickups_per_avatar',
            field=models.FloatField(default=1.2),
        ),
    ]



from django.core.exceptions import ValidationError
from django.forms import ModelForm

from aimmo.models import Game


class AddGameForm(ModelForm):
    def __init__(self, playable_games, *args, **kwargs):
        self.playable_games = playable_games
        super(AddGameForm, self).__init__(*args, **kwargs)

    class Meta:
        model = Game
        exclude = ['Main', 'owner', 'auth_token', 'completed', 'main_user', 'static_data', 'can_play',
                   'public', 'generator', 'target_num_cells_per_avatar',
                   'target_num_score_locations_per_avatar', 'score_despawn_chance',
                   'target_num_pickups_per_avatar', 'pickup_spawn_chance', 'obstacle_ratio',
                   'start_height', 'start_width'
                   ]

    def clean(self):
        name = self.cleaned_data['name']

        playable_games_names = [
            playable_game.name
            for playable_game in self.playable_games
            if self.playable_games and name
        ]

        if name in playable_games_names:
            raise ValidationError("Sorry, a game with this name already exists.")

        return self.cleaned_data



import math
from logging import getLogger

from simulation.level_settings import DEFAULT_LEVEL_SETTINGS
from simulation.location import Location
from simulation.game_logic import SpawnLocationFinder, ScoreLocationUpdater, MapContext, PickupUpdater, MapExpander
from simulation.cell import Cell

LOGGER = getLogger(__name__)


class WorldMap(object):
    """
    The non-player world state.
    """

    def __init__(self, grid, settings):
        """
        :param grid: All types of cells to be inserted into the map.
        :param settings: Constant values provided when generating a level/map.
        """
        self.grid = grid
        self.settings = settings
        self._spawn_location_finder = SpawnLocationFinder(self)

    @classmethod
    def _min_max_from_dimensions(cls, height, width):
        """
        The value provided by the user will be an integer both for the width and height
        components. We calculate the maximum and minimum dimensions in all directions.
        """
        max_x = int(math.floor(width / 2))
        min_x = -(width - max_x - 1)
        max_y = int(math.floor(height / 2))
        min_y = -(height - max_y - 1)
        return min_x, max_x, min_y, max_y

    @classmethod
    def generate_empty_map(cls, height, width, settings):
        new_settings = DEFAULT_LEVEL_SETTINGS.copy()
        new_settings.update(settings)

        (min_x, max_x, min_y, max_y) = WorldMap._min_max_from_dimensions(height, width)
        grid = {}
        for x in range(min_x, max_x + 1):
            for y in range(min_y, max_y + 1):
                location = Location(x, y)
                grid[location] = Cell(location)
        return cls(grid, new_settings)

    def all_cells(self):
        return self.grid.values()

    def score_cells(self):
        return (c for c in self.all_cells() if c.generates_score)

    def pickup_cells(self):
        return (c for c in self.all_cells() if c.pickup)

    def is_on_map(self, location):
        try:
            self.grid[location]
        except KeyError:
            return False
        return True

    def get_cell(self, location):
        try:
            return self.grid[location]
        except KeyError:
            # For backwards-compatibility, this throws ValueError
            raise ValueError('Location %s is not on the map' % location)

    def get_cell_by_coords(self, x, y):
        return self.get_cell(Location(x, y))

    def clear_cell_actions(self, location):
        try:
            cell = self.get_cell(location)
            cell.actions = []
        except ValueError:
            return

    def max_y(self):
        return max(self.grid.keys(), key=lambda c: c.y).y

    def min_y(self):
        return min(self.grid.keys(), key=lambda c: c.y).y

    def max_x(self):
        return max(self.grid.keys(), key=lambda c: c.x).x

    def min_x(self):
        return min(self.grid.keys(), key=lambda c: c.x).x

    @property
    def num_rows(self):
        return self.max_y() - self.min_y() + 1

    @property
    def num_cols(self):
        return self.max_x() - self.min_x() + 1

    @property
    def num_cells(self):
        return self.num_rows * self.num_cols

    def update(self, num_avatars):
        self._update_avatars()
        self._update_map(num_avatars)

    def _update_avatars(self):
        self._apply_score()
        self._apply_pickups()

    def _apply_pickups(self):
        for cell in self.pickup_cells():
            if cell.avatar is not None:
                cell.pickup.apply(cell.avatar)

    def _apply_score(self):
        for cell in self.score_cells():
            try:
                cell.avatar.score += 1
            except AttributeError:
                pass

    def _update_map(self, num_avatars):
        context = MapContext(num_avatars=num_avatars)
        MapExpander().update(self, context=context)
        ScoreLocationUpdater().update(self, context=context)
        PickupUpdater().update(self, context=context)

    def can_move_to(self, target_location):
        if not self.is_on_map(target_location):
            return False
        cell = self.get_cell(target_location)

        return (cell.habitable
                and (not cell.is_occupied or cell.avatar.is_moving)
                and len(cell.moves) <= 1)

    def attackable_avatar(self, target_location):
        """
        Return a boolean if the avatar is attackable at the given location (or will be
        after next move), else return None.
        """
        try:
            cell = self.get_cell(target_location)
        except ValueError:
            return None

        if cell.avatar:
            return cell.avatar

        if len(cell.moves) == 1:
            return cell.moves[0].avatar

        return None

    def get_no_fog_distance(self):
        return self.settings['NO_FOG_OF_WAR_DISTANCE']

    def get_partial_fog_distance(self):
        return self.settings['PARTIAL_FOG_OF_WAR_DISTANCE']

    def get_random_spawn_location(self):
        return self._spawn_location_finder.get_random_spawn_location()

    def __repr__(self):
        return repr(self.grid)

    def __iter__(self):
        return ((self.get_cell(Location(x, y))
                for y in range(self.min_y(), self.max_y() + 1))
                for x in range(self.min_x(), self.max_x() + 1))

    # Serialisation Utilities
    def get_serialised_south_west_corner(self):
        """
        Used in serialising the map size when sent to the front end. Very lightweight as
        it consists of two integers.

        :return: A dictionary with two values, x and y coordinates for the bottom left
        (south-west) corner of the map.
        """
        return {
            "x": self.min_x(),
            "y": self.min_y(),
        }

    def get_serialised_north_east_corner(self):
        """
        Used in serialising the map size when sent to the front end. Very lightweight as
        it consists of two integers.

        :return: A dictionary with two values, x and y coordinates for the top right
        (north-west) corner of the map.
        """
        return {
            "x": self.max_x(),
            "y": self.max_y(),
        }

    def serialise_score_location(self):
        """
        Used to serialise the score locations on every update.

        :return: A single list that contains all score locations. Within
        the list there are x and y coordinates.
        """

        def get_coords(cell):
            return {'location': {'x': cell.location.x, 'y': cell.location.y}}

        return [get_coords(cell) for cell in self.all_cells() if cell.generates_score]

    def serialise_obstacles(self):
        """
        Used to serialise the obstacle locations on every update.

        :return: A list that contains all the obstacle information generated by inner method.
        """

        def serialise_obstacle(cell):
            return {'location': {
                         'x': cell.location.x,
                         'y': cell.location.y,
                     },
                    'width': 1,
                    'height': 1,
                    'type': "wall",
                    'orientation': "north",
                    }

        return [serialise_obstacle(cell) for cell in self.all_cells() if not cell.habitable]


def WorldMapStaticSpawnDecorator(world_map, spawn_location):
    world_map._spawn_location_finder.get_random_spawn_location = lambda: spawn_location
    return world_map



import logging
import time

from eventlet.semaphore import Semaphore
from threading import Thread
from concurrent import futures

from simulation.worker import Worker

LOGGER = logging.getLogger(__name__)


class WorkerManager(object):
    """
    Methods of this class must be thread safe unless explicitly stated.
    """
    def __init__(self, port=5000):
        self.player_id_to_worker = {}
        self.port = port

    def get_code(self, player_id):
        return self.player_id_to_worker[player_id].code

    def fetch_all_worker_data(self, player_id_to_game_state):
        """
        Creates a thread for each worker to send a request for their data. After
        a set duration these threads will close, giving a consistent turn time.
        """
        def prepare_request_threads():
            return [Thread(target=worker.fetch_data,
                           args=(player_id_to_game_state[player_id],))
                    for (player_id, worker) in self.player_id_to_worker.items()]

        def timed_process_for_worker_turn_requests(duration):
            threads = prepare_request_threads()

            [thread.setDaemon(True) for thread in threads]
            [thread.start() for thread in threads]
            time.sleep(duration)

        timed_process_for_worker_turn_requests(2)

    def get_player_id_to_serialised_actions(self):
        return {player_id: self.player_id_to_worker[player_id].serialised_action for player_id in self.player_id_to_worker}

    def clear_logs(self):
        for worker in self.player_id_to_worker.values():
            worker.log = None

    def create_worker(self, player_id):
        raise NotImplementedError

    def remove_worker(self, player_id):
        raise NotImplementedError

    def update_code(self, player):
        self.player_id_to_worker[player['id']].code = player['code']

    def add_new_worker(self, player_id):
        worker_url_base = self.create_worker(player_id)
        self.player_id_to_worker[player_id] = Worker(f'{worker_url_base}/turn/')

    def _parallel_map(self, func, iterable_args):
        with futures.ThreadPoolExecutor() as executor:
            results = executor.map(func, iterable_args)

    def add_workers(self, users_to_add):
        self._parallel_map(self.add_new_worker, users_to_add)

    def delete_workers(self, players_to_delete):
        self._parallel_map(self.delete_worker, players_to_delete)

    def delete_worker(self, player):
        del self.player_id_to_worker[player]
        self.remove_worker(player)

    def update_worker_codes(self, players):
        self._parallel_map(self.update_code, players)



class MockCommunicator(object):
    def __init__(self):
        self.data = {
            "main": {
                "parameters": [],
                "main_avatar": 1,
                "users": [
                    {
                        "id": 1,
                        "code": "class Avatar(object):\n"
                                "    def handle_turn(self, world_view, events):\n"
                                "        from simulation.action import WaitAction\n"
                                "        return WaitAction()\n"
                    },

                    {
                        "id": 2,
                        "code": "class Avatar(object):\n"
                                "    def handle_turn(self, world_view, events):\n"
                                "        from simulation.action import WaitAction\n"
                                "        return WaitAction()\n"
                    },
                ]
            }
        }

    def get_game_metadata(self):
        return self.data

    def mark_game_complete(self):
        return {}

    def change_code(self, avatar_id, new_code):
        users = self.data['main']['users']
        for i in range(len(users)):
            if users[i]['id'] == avatar_id:
                users[i]['code'] = new_code



from threading import RLock
from simulation.pickups import serialise_pickups


class GameState(object):
    """
    Encapsulates the entire game state, including avatars, their code, and the world.
    """

    def __init__(self, world_map, avatar_manager, completion_check_callback=lambda: None):
        self.world_map = world_map
        self.avatar_manager = avatar_manager
        self._completion_callback = completion_check_callback
        self.main_avatar_id = None
        self._lock = RLock()

    def add_avatar(self, player_id, location=None):
        with self._lock:
            location = self.world_map.get_random_spawn_location() if location is None else location
            avatar = self.avatar_manager.add_avatar(player_id, location)
            self.world_map.get_cell(location).avatar = avatar

    def add_avatars(self, player_ids):
        for player_id in player_ids:
            self.add_avatar(player_id)

    def delete_avatars(self, player_ids):
        for player_id in player_ids:
            self.remove_avatar(player_id)

    def remove_avatar(self, player_id):
        with self._lock:
            try:
                avatar = self.avatar_manager.get_avatar(player_id)
            except KeyError:
                return
            self.world_map.get_cell(avatar.location).avatar = None
            self.avatar_manager.remove_avatar(player_id)

    def _update_effects(self):
        with self._lock:
            for avatar in self.avatar_manager.active_avatars:
                avatar.update_effects()

    def update_environment(self):
        with self._lock:
            self._update_effects()
            num_avatars = len(self.avatar_manager.active_avatars)
            self.world_map.update(num_avatars)

    def is_complete(self):
        with self._lock:
            return self._completion_callback(self)

    def get_main_avatar(self):
        with self._lock:
            return self.avatar_manager.avatars_by_id[self.main_avatar_id]

    def serialise(self):
        return {
            'era': "less_flat",
            'southWestCorner': self.world_map.get_serialised_south_west_corner(),
            'northEastCorner': self.world_map.get_serialised_north_east_corner(),
            'players': self.avatar_manager.serialise_players(),
            'pickups': serialise_pickups(self.world_map),
            'scoreLocations': (self.world_map.serialise_score_location()),
            'obstacles': self.world_map.serialise_obstacles()
        }

    def serialise_for_worker(self, avatar_wrapper):
        with self._lock:
            return {
                'avatar_state': avatar_wrapper.serialise(),
                'world_map': {
                    'cells': [cell.serialise() for cell in self.world_map.all_cells()]
                }
            }

    def get_serialised_game_states_for_workers(self):
        with self._lock:
            return {player_id: self.serialise_for_worker(avatar_wrapper) for player_id, avatar_wrapper
                    in self.avatar_manager.avatars_by_id.items()}



import logging
import asyncio
from abc import ABCMeta, abstractmethod
from concurrent.futures import ALL_COMPLETED, ThreadPoolExecutor
from simulation.action import PRIORITIES, WaitAction
from threading import Thread

LOGGER = logging.getLogger(__name__)

TURN_INTERVAL = 2


class SimulationRunner(object):
    """
    Game loop
    """
    daemon = True
    __metaclass__ = ABCMeta

    def __init__(self, game_state, communicator):
        self.game_state = game_state
        self.communicator = communicator

    @abstractmethod
    async def run_turn(self, player_id_to_serialised_actions):
        pass

    async def _run_turn_for_avatar(self, avatar, serialised_action):
        """
        Send an avatar its view of the game state and register its
        chosen action & logs.
        """
        self._register_actions(avatar, serialised_action)

    def _register_actions(self, avatar, serialised_action):
        """
        Calls a function that constructs the action object, does error handling,
        and finally registers it onto the avatar.

        :param avatar: Avatar wrapper object
        :param serialised_action: A string representing the action
        """
        if avatar.decide_action(serialised_action):
            avatar.action.register(self.game_state.world_map)

    def _update_environment(self, game_state):
        num_avatars = len(game_state.avatar_manager.active_avatars)
        game_state.world_map.reconstruct_interactive_state(num_avatars)

    def _mark_complete(self):
        self.communicator.mark_game_complete(data=self.game_state.serialise())

    async def run_single_turn(self, player_id_to_serialised_actions):
        await self.run_turn(player_id_to_serialised_actions)
        self.game_state.update_environment()


class SequentialSimulationRunner(SimulationRunner):
    async def run_turn(self, player_id_to_serialised_actions):
        """
        Get and apply each avatar's action in turn.
        """
        avatars = self.game_state.avatar_manager.active_avatars

        for avatar in avatars:
            await self._run_turn_for_avatar(avatar, player_id_to_serialised_actions[avatar.player_id])
            location_to_clear = avatar.action.target_location
            avatar.action.process(self.game_state.world_map)
            self.game_state.world_map.clear_cell_actions(location_to_clear)


class ConcurrentSimulationRunner(SimulationRunner):
    async def async_map(self, func, iterable_args):
        futures = [func(*arg) for arg in iterable_args]
        await asyncio.gather(*futures)

    async def run_turn(self, player_id_to_serialised_actions):
        """
        Concurrently get the intended actions from all avatars and register
        them on the world map. Then apply actions in order of priority.
        """

        avatars = self.game_state.avatar_manager.active_avatars
        args = [(avatar, player_id_to_serialised_actions[avatar.player_id]) for avatar in avatars]
        await self.async_map(self._run_turn_for_avatar, args)

        # Waits applied first, then attacks, then moves.
        avatars.sort(key=lambda a: PRIORITIES[type(a.action)])

        locations_to_clear = {a.action.target_location for a in avatars
                              if a.action is not None}

        for action in (a.action for a in avatars if a.action is not None):
            action.process(self.game_state.world_map)

        for location in locations_to_clear:
            self.game_state.world_map.clear_cell_actions(location)



# -*- coding: utf-8 -*-
from setuptools import find_packages, setup

import versioneer

setup(
    name='aimmo',
    cmdclass=versioneer.get_cmdclass(),
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'django >= 1.8.3, <= 1.10.8',
        'django-autoconfig >= 0.3.6, < 1.0.0',
        'django-forms-bootstrap',
        'django-js-reverse',
        'eventlet',
        'flask',
        'flask-socketio',
        'requests',
        'six',
        'hypothesis',
        'flask-cors >= 3.0, < 3.1',
        'psutil >= 5.4, < 5.5',
        'RestrictedPython == 4.0.b7'
    ],
    tests_require=[
        'httmock',
        'mock == 2.0.0',
        'docker >= 3.5, < 3.6',
        'kubernetes == 5.0.0',
        'PyYAML == 3.12',
    ],

    version=versioneer.get_version(),
    zip_safe=False,
)



from abc import ABCMeta, abstractmethod, abstractproperty
import simulation.effects as effects

DAMAGE_BOOST_DEFAULT = 5
HEALTH_RESTORE_DEFAULT = 3
HEALTH_RESTORE_MAX = 100
AVATAR_HEALTH_MAX = 100


class _Pickup(object):
    __metaclass__ = ABCMeta

    def __init__(self, cell):
        self.cell = cell

    def __str__(self):
        return self.__class__.__name__

    def delete(self):
        self.cell.pickup = None

    def apply(self, avatar):
        self._apply(avatar)
        self.delete()

    @abstractmethod
    def _apply(self, avatar):
        raise NotImplementedError()

    @abstractmethod
    def serialise(self):
        raise NotImplementedError()


class HealthPickup(_Pickup):
    def __init__(self, cell, health_restored=HEALTH_RESTORE_DEFAULT):
        # Round the integer up to the nearest value (ceiling).
        health_restored = int(round(health_restored))

        # Check if the value provided is legal.
        if 0 < health_restored <= HEALTH_RESTORE_MAX:
            super(HealthPickup, self).__init__(cell)
            self.health_restored = health_restored
        else:
            raise ValueError("Health Restored has to be within 0-100 range!")

    def __repr__(self):
        return 'HealthPickup(health_restored={})'.format(self.health_restored)

    def serialise(self):
        return {
                'type': 'health',
                'location': {
                    'x': self.cell.location.x,
                    'y': self.cell.location.y,
                }
        }

    def _apply(self, avatar):
        avatar.health += self.health_restored

        # Make sure the health is capped at 100.
        if avatar.health > AVATAR_HEALTH_MAX:
            avatar.health = AVATAR_HEALTH_MAX


class _PickupEffect(_Pickup):
    __metaclass__ = ABCMeta

    def __init__(self, *args):
        super(_PickupEffect, self).__init__(*args)
        self.params = []

    @abstractproperty
    def EFFECT(self):
        raise NotImplementedError()

    def _apply(self, avatar):
        self.params.append(avatar)
        avatar.effects.add(self.EFFECT(*self.params))


class InvulnerabilityPickup(_PickupEffect):
    EFFECT = effects.InvulnerabilityPickupEffect

    def serialise(self):
        return {
                'type': 'invulnerability',
                'location': {
                    'x': self.cell.location.x,
                    'y': self.cell.location.y,
                }
        }


class DamageBoostPickup(_PickupEffect):
    EFFECT = effects.DamageBoostPickupEffect

    def __init__(self, cell, damage_boost=DAMAGE_BOOST_DEFAULT):
        if damage_boost <= 0:
            raise ValueError("The damage_boost parameter is less than or equal to 0!")

        super(DamageBoostPickup, self).__init__(cell)
        self.damage_boost = damage_boost
        self.params.append(self.damage_boost)

    def __repr__(self):
        return 'DamageBoostPickup(damage_boost={})'.format(self.damage_boost)

    def serialise(self):
        return {
                'type': 'damage_boost',
                'location': {
                    'x': self.cell.location.x,
                    'y': self.cell.location.y,
                }
        }


def serialise_pickups(world_map):
    return [cell.pickup.serialise() for cell in world_map.pickup_cells()]


ALL_PICKUPS = (
    HealthPickup,
    InvulnerabilityPickup,
    DamageBoostPickup,
)



import os
import mock

from unittest import TestCase
from urllib.parse import urlparse

from simulation.worker_managers.local_worker_manager import LocalWorkerManager


class TestLocalWorkerManager(TestCase):
    @mock.patch('docker.from_env')
    def test_local_worker_ports_do_not_conflict(self, docker_from_env):
        os.environ['GAME_ID'] = '1'
        worker_manager1 = LocalWorkerManager()
        local_worker1 = worker_manager1.create_worker(1)
        url1 = urlparse(local_worker1)

        os.environ['GAME_ID'] = '2'
        worker_manager2 = LocalWorkerManager()
        local_worker2 = worker_manager2.create_worker(1)
        url2 = urlparse(local_worker2)

        self.assertEqual(url1.port, 11989)
        self.assertEqual(url2.port, 21989)

    @mock.patch('docker.from_env')
    def test_local_workers_in_the_same_game_do_not_have_port_conflicts(self, docker_from_env):
        os.environ['GAME_ID'] = '1'
        worker_manager = LocalWorkerManager()
        local_worker1 = worker_manager.create_worker(1)
        local_worker2 = worker_manager.create_worker(2)
        url1 = urlparse(local_worker1)
        url2 = urlparse(local_worker2)

        self.assertEqual(url1.port, 11989)
        self.assertEqual(url2.port, 11990)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models, transaction, IntegrityError

import aimmo.models


def migrate_data_forward(apps, schema_editor):
    Player = apps.get_model("aimmo", "Player")
    Avatar = apps.get_model("aimmo", "Avatar")
    Game = apps.get_model("aimmo", "Game")

    if Player.objects.count() == 0:
        return
    main_game = Game(pk=1, name="main")
    main_game.save()

    avatars = [Avatar(game=main_game, owner=player.user, code=player.code)
               for player in Player.objects.all()]
    Avatar.objects.bulk_create(avatars)


def migrate_data_backward(apps, schema_editor):
    Player = apps.get_model("aimmo", "Player")
    Avatar = apps.get_model("aimmo", "Avatar")

    if Avatar.objects.count() == 0:
        return
    for avatar in Avatar.objects.all():
        player = Player(user=avatar.owner, code=avatar.code)
        try:
            with transaction.atomic():
                player.save()
        except IntegrityError:
            # Stuff doesn't map as can have more than one Avatar but only one Player
            pass


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('aimmo', '0002_auto_20160601_1914'),
    ]

    operations = [
        migrations.CreateModel(
            name='Avatar',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('code', models.TextField()),
                ('auth_token', models.CharField(default=aimmo.models.generate_auth_token, max_length=24)),
            ],
        ),
        migrations.CreateModel(
            name='Game',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=100)),
                ('auth_token', models.CharField(default=aimmo.models.generate_auth_token, max_length=24)),
                ('public', models.BooleanField(default=True)),
                ('can_play', models.ManyToManyField(related_name='playable_games', to=settings.AUTH_USER_MODEL)),
                ('owner', models.ForeignKey(related_name='owned_games', to=settings.AUTH_USER_MODEL, null=True)),
            ],
        ),
        migrations.AddField(
            model_name='avatar',
            name='game',
            field=models.ForeignKey(to='aimmo.Game'),
        ),
        migrations.AddField(
            model_name='avatar',
            name='owner',
            field=models.ForeignKey(to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterUniqueTogether(
            name='avatar',
            unique_together=set([('owner', 'game')]),
        ),
        migrations.RunPython(
            migrate_data_forward,
            migrate_data_backward
        ),
        migrations.RemoveField(
            model_name='player',
            name='user',
        ),
        migrations.DeleteModel(
            name='Player',
        ),
    ]



from __future__ import absolute_import

import unittest
import asyncio

from simulation.avatar.avatar_appearance import AvatarAppearance
from simulation.game_state import GameState
from simulation.location import Location
from simulation.simulation_runner import ConcurrentSimulationRunner

from .dummy_avatar import (DummyAvatarManager, MoveEastDummy, MoveNorthDummy,
                           MoveSouthDummy, MoveWestDummy, WaitDummy, DeadDummy)
from .maps import InfiniteMap
from .mock_communicator import MockCommunicator

ORIGIN = Location(0, 0)

RIGHT_OF_ORIGIN = Location(1, 0)
FIVE_RIGHT_OF_ORIGIN = Location(5, 0)

ABOVE_ORIGIN = Location(0, 1)
FIVE_RIGHT_OF_ORIGIN_AND_ONE_ABOVE = Location(5, 1)


class MockGameState(GameState):
    def get_state_for(self, avatar):
        return self


class TestSimulationRunner(unittest.TestCase):
    """
        Key:
            > : Avatar moving eastward
            < : Avatar moving westward
            x : Avatar waiting / blocked
            o : Avatar successfully moved
            ! : Dead avatar (that should be waiting)
    """

    def construct_default_avatar_appearance(self):
        return AvatarAppearance("#000", "#ddd", "#777", "#fff")

    def construct_simulation_runner(self, avatars, locations):
        self.avatar_manager = DummyAvatarManager(avatars)
        self.avatar_manager.avatars_by_id = dict(enumerate(avatars))
        self.game_state = MockGameState(InfiniteMap(), self.avatar_manager)
        self.simulation_runner = ConcurrentSimulationRunner(game_state=self.game_state,
                                                            communicator=MockCommunicator())
        for index, location in enumerate(locations):
            self.game_state.add_avatar(index, location)

    def assert_at(self, avatar, location):
        self.assertEqual(avatar.location, location)
        cell = self.game_state.world_map.get_cell(location)
        self.assertEqual(cell.avatar, avatar)

    def get_avatar(self, player_id):
        return self.avatar_manager.get_avatar(player_id)

    def run_turn(self):
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self.simulation_runner.run_turn(self.avatar_manager.avatars_by_id))

    def test_run_turn(self):
        """
        Given:  > _
        (1)
        Expect: _ o
        """
        self.construct_simulation_runner([MoveEastDummy], [ORIGIN])
        avatar = self.get_avatar(0)

        self.assert_at(avatar, ORIGIN)
        self.run_turn()
        self.assert_at(avatar, RIGHT_OF_ORIGIN)

    def test_run_several_turns(self):
        """
        Given:  > _ _ _ _ _
        (5)
        Expect: _ _ _ _ _ o
        """
        self.construct_simulation_runner([MoveEastDummy], [ORIGIN])
        avatar = self.get_avatar(0)

        self.assertEqual(avatar.location, ORIGIN)
        [self.run_turn() for _ in range(5)]
        self.assertEqual(avatar.location, FIVE_RIGHT_OF_ORIGIN)

    def test_run_several_turns_and_avatars(self):
        """
        Given:  > _ _ _ _ _
                > _ _ _ _ _
        (5)
        Expect: _ _ _ _ _ o
                _ _ _ _ _ o
        """
        self.construct_simulation_runner([MoveEastDummy, MoveEastDummy],
                                    [ORIGIN,        ABOVE_ORIGIN])
        avatar0 = self.get_avatar(0)
        avatar1 = self.get_avatar(1)

        self.assert_at(avatar0, ORIGIN)
        self.assert_at(avatar1, ABOVE_ORIGIN)
        [self.run_turn() for _ in range(5)]
        self.assert_at(avatar0, FIVE_RIGHT_OF_ORIGIN)
        self.assert_at(avatar1, FIVE_RIGHT_OF_ORIGIN_AND_ONE_ABOVE)

    def test_move_chain_succeeds(self):
        """
        Given:  > > > > > _

        Expect: _ o o o o o
        """
        self.construct_simulation_runner([MoveEastDummy for _ in range(5)],
                                    [Location(x, 0) for x in range(5)])
        avatars = [self.get_avatar(i) for i in range(5)]

        [self.assert_at(avatars[x], Location(x, 0)) for x in range(5)]
        self.run_turn()
        [self.assert_at(avatars[x], Location(x + 1, 0)) for x in range(5)]

    def test_move_chain_fails_occupied(self):
        """
        Given:  > > x _

        Expect: x x x _
        """
        self.construct_simulation_runner([MoveEastDummy, MoveEastDummy, WaitDummy],
                                    [Location(x, 0) for x in range(3)])
        avatars = [self.get_avatar(i) for i in range(3)]

        [self.assert_at(avatars[x], Location(x, 0)) for x in range(3)]
        self.run_turn()
        [self.assert_at(avatars[x], Location(x, 0)) for x in range(3)]

    def test_move_chain_fails_occupied_by_dead_avatar(self):
        """
        Given: > > ! _

        Expect: x x ! _
        """

        self.construct_simulation_runner([MoveEastDummy, MoveEastDummy, DeadDummy], [Location(x, 0) for x in range(3)])
        avatars = [self.get_avatar(i) for i in range(3)]

        [self.assert_at(avatars[x], Location(x, 0)) for x in range(3)]
        self.run_turn()
        [self.assert_at(avatars[x], Location(x, 0)) for x in range(3)]

    def test_move_fails_collision(self):
        """
        Given: > _ <
        Expect: x _ x
        """

        self.construct_simulation_runner([MoveEastDummy, MoveWestDummy], [Location(0, 0), Location(2, 0)])
        avatars = [self.get_avatar(i) for i in range(2)]

        self.assert_at(avatars[0], Location(0, 0))
        self.assert_at(avatars[1], Location(2, 0))

        self.run_turn()

        self.assert_at(avatars[0], Location(0, 0))
        self.assert_at(avatars[1], Location(2, 0))

    def test_move_chain_fails_collision(self):
        """
        Given:  > > > _ <
        (1)
        Expect: x x x _ x
        """
        locations = [Location(0, 0), Location(1, 0), Location(2, 0), Location(4, 0)]
        self.construct_simulation_runner(
            [MoveEastDummy, MoveEastDummy, MoveEastDummy, MoveWestDummy],
            locations)
        avatars = [self.get_avatar(i) for i in range(4)]

        [self.assert_at(avatars[i], locations[i]) for i in range(4)]
        self.run_turn()
        [self.assert_at(avatars[i], locations[i]) for i in range(4)]

    def test_move_chain_fails_cycle(self):
        """
        Given:  > v
                ^ <
        (1)
        Expect: x x
                x x
        """
        locations = [Location(0, 1), Location(1, 1), Location(1, 0), Location(0, 0)]
        self.construct_simulation_runner(
            [MoveEastDummy, MoveSouthDummy, MoveWestDummy, MoveNorthDummy],
            locations)
        avatars = [self.get_avatar(i) for i in range(4)]

        [self.assert_at(avatars[i], locations[i]) for i in range(4)]
        self.run_turn()
        [self.assert_at(avatars[i], locations[i]) for i in range(4)]

    def test_move_chain_fails_spiral(self):
        """
        Given:  > > v
                  ^ <
        (1)
        Expect: x x x
                  x x
        """
        locations = [Location(0, 1),
                     Location(1, 1),
                     Location(2, 1),
                     Location(2, 0),
                     Location(1, 0)]
        self.construct_simulation_runner(
            [MoveEastDummy, MoveEastDummy, MoveSouthDummy, MoveWestDummy, MoveNorthDummy],
            locations)
        avatars = [self.get_avatar(i) for i in range(5)]

        [self.assert_at(avatars[i], locations[i]) for i in range(5)]
        self.run_turn()
        [self.assert_at(avatars[i], locations[i]) for i in range(5)]


if __name__ == '__main__':
    unittest.main()



from django.contrib.auth.models import User
from django.test import TestCase

from aimmo import models, forms, app_settings

app_settings.GAME_SERVER_URL_FUNCTION = lambda game_id: ('base %s' % game_id, 'path %s' % game_id)
app_settings.GAME_SERVER_PORT_FUNCTION = lambda game_id: 0
app_settings.GAME_SERVER_SSL_FLAG = True


class TestForms(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user1 = User.objects.create_user('test', 'test@example.com', 'password')
        cls.user1.save()
        cls.game = models.Game(id=1, name='test', public=False)
        cls.game.save()

    def test_create_game(self):
        form = forms.AddGameForm([self.game], {'name': 'test2'})
        self.assertTrue(form.is_valid())
        new_game = form.save()
        self.assertEqual(new_game.name, "test2")

    def test_cannot_create_duplicate_game(self):
        form = forms.AddGameForm([self.game], {'name': 'test'})
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors, {'__all__': [u'Sorry, a game with this name already exists.']})



from __future__ import absolute_import

from unittest import TestCase

from simulation.game_state import GameState
from simulation.location import Location
from .dummy_avatar import DummyAvatar
from .dummy_avatar import DummyAvatarManager
from .maps import InfiniteMap, AvatarMap, EmptyMap


class TestGameState(TestCase):
    def test_remove_non_existant_avatar(self):
        state = GameState(None, DummyAvatarManager())
        state.remove_avatar(10)

    def test_remove_avatar(self):
        world_map = InfiniteMap()
        manager = DummyAvatarManager()
        game_state = GameState(world_map, manager)

        avatar1 = DummyAvatar(1, Location(0, 0))
        avatar2 = DummyAvatar(2, Location(1, 1))
        avatar2.marked = True

        manager.add_avatar_directly(avatar1)
        world_map.get_cell(Location(0, 0)).avatar = avatar1
        manager.add_avatar_directly(avatar2)
        world_map.get_cell(Location(1, 1)).avatar = avatar2

        game_state.remove_avatar(1)

        self.assertNotIn(1, manager.avatars_by_id)
        self.assertEqual(world_map.get_cell(Location(0, 0)).avatar, None)

        self.assertTrue(manager.avatars_by_id[2].marked)
        self.assertTrue(world_map.get_cell(Location(1, 1)).avatar.marked)

    def game_state_with_two_avatars(self, world_map=None, avatar_manager=None):
        if world_map is None:
            world_map = EmptyMap()
        if avatar_manager is None:
            avatar_manager = DummyAvatarManager()

        avatar = DummyAvatar(1, (0, 0))
        other_avatar = DummyAvatar(2, (0, 0))
        other_avatar.marked = True
        avatar_manager.avatars_by_id[1] = avatar
        avatar_manager.avatars_by_id[2] = other_avatar
        game_state = GameState(world_map, avatar_manager)

        return (game_state, avatar, world_map, avatar_manager)

    def test_add_avatar(self):
        state = GameState(AvatarMap(None), DummyAvatarManager())
        state.add_avatar(7)
        self.assertIn(7, state.avatar_manager.avatars_by_id)
        avatar = state.avatar_manager.avatars_by_id[7]
        self.assertEqual(avatar.location.x, 10)
        self.assertEqual(avatar.location.y, 10)

    def test_updates_map(self):
        map = InfiniteMap()
        state = GameState(map, DummyAvatarManager())
        state.update_environment()
        self.assertEqual(map.updates, 1)

    def test_updates_map_with_correct_num_avatars(self):
        map = InfiniteMap()
        manager = DummyAvatarManager()
        manager.add_avatar(1)
        state = GameState(map, manager)
        state.update_environment()
        self.assertEqual(map.num_avatars, 1)
        manager.add_avatar(2)
        manager.add_avatar(3)
        state.update_environment()
        self.assertEqual(map.num_avatars, 3)

    def test_no_main_avatar_by_default(self):
        state = GameState(EmptyMap(), DummyAvatarManager())
        with self.assertRaises(KeyError):
            state.get_main_avatar()

    def test_get_main_avatar(self):
        (game_state, avatar, _, _) = self.game_state_with_two_avatars()
        game_state.main_avatar_id = avatar.player_id
        self.assertEqual(game_state.get_main_avatar(), avatar)

    def test_is_complete_calls_lambda(self):
        class LambdaTest(object):
            def __init__(self, return_value):
                self.return_value = return_value

            def __call__(self, game_state):
                self.game_state = game_state
                return self.return_value

        test = LambdaTest(True)
        game_state = GameState(EmptyMap(), DummyAvatarManager(), test)
        self.assertTrue(game_state.is_complete())
        test.return_value = False
        self.assertFalse(game_state.is_complete())
